package daemon

import (
	"context"
	"flag"
	"fmt"
	"math/rand"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
)

var log = _log.Default
var DefaultInterval = 5 * time.Second

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&DefaultInterval, "interval", DefaultInterval, "Default module execution interval duration")
}

type DaemonFactoryI interface {
	Daemons(inventory *common.Inventory) map[string]interface{}
	New(params interface{}) DaemonI
}

// For factories with configuration files
type ReloaderDaemonFactoryI interface {
	Reload() error
}

type DaemonI interface {
	Interval() time.Duration
	Delay() time.Duration
	Init(context.Context) error
	Collect(context.Context, *common.MetricData) error
	Close() error
}

type BaseDaemon struct{}

func (d *BaseDaemon) Interval() time.Duration {
	return DefaultInterval
}

func (d *BaseDaemon) Delay() time.Duration {
	// Sleep for a random amount of time in [0;$interval) to spread the load
	interval := d.Interval()
	if interval == 0 {
		return 0
	}
	return time.Duration(rand.Int63n(int64(interval)))
}

func (d *BaseDaemon) Init(ctx context.Context) error {
	return nil
}

func (d *BaseDaemon) Close() error {
	return nil
}

func Run(ctx context.Context, daemonID string, d DaemonI, m *common.MetricData) error {
	doneC := ctx.Done()

	initRetryAfter := 5 * time.Minute
	for {
		if err := d.Init(ctx); err != nil {
			log.Warnf("Init error for daemon '%s', retrying in %s: %s", daemonID, initRetryAfter, err)

			select {
			case <-doneC:
				return close(daemonID, d)
			case <-time.After(initRetryAfter):
			}

			continue
		}

		break
	}

	interval := d.Interval()
	if interval <= 0 {
		interval = DefaultInterval
	}

	timer := time.AfterFunc(time.Hour, func() {
		log.Warnf("%s collection is taking longer than its interval of %s", daemonID, interval)
	})
	ticker := time.NewTicker(interval)
	defer ticker.Stop()

	// In case of a long init
	select {
	case <-doneC:
		return close(daemonID, d)
	default:
	}

	timer.Reset(interval)
	// TODO: add deadline to context?
	if err := d.Collect(ctx, m); err != nil {
		log.Errorf("error collecting %s: %s", daemonID, err)
	}
	timer.Stop()

	for {
		// This is required
		// If ticker.C is already ready by the time the loop starts again
		// The second select's behavior wouldn't be determined
		select {
		case <-doneC:
			return close(daemonID, d)
		default:
		}

		select {
		case <-doneC:
			return close(daemonID, d)
		case <-ticker.C:
		}

		timer.Reset(interval)
		// TODO: add deadline to context?
		if err := d.Collect(ctx, m); err != nil {
			log.Errorf("error collecting %s: %s", daemonID, err)
		}
		timer.Stop()
	}
}

func close(daemonID string, d DaemonI) error {
	if err := d.Close(); err != nil {
		return fmt.Errorf("Closing %s: %w", daemonID, err)
	}

	return nil
}
