package s3rt

import (
	"bytes"
	"context"
	"crypto/md5"
	"flag"
	"fmt"
	"io"
	"io/ioutil"
	"net/http"
	"os"
	"strconv"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"

	"github.com/aws/aws-sdk-go/aws"
	"github.com/aws/aws-sdk-go/aws/awserr"
	awsRequest "github.com/aws/aws-sdk-go/aws/request"
	"github.com/aws/aws-sdk-go/aws/session"
	"github.com/aws/aws-sdk-go/service/s3"
	"github.com/aws/aws-sdk-go/service/s3/s3iface"
	"github.com/aws/aws-sdk-go/service/s3/s3manager"
	"github.com/aws/aws-sdk-go/service/s3/s3manager/s3manageriface"
)

var log = _log.New("s3rt")

var (
	ramBuffers   = true
	mpuSize      = 5 * 1024 * 1024
	size         = mpuSize - 1
	bucketRT     = false
	confTimeout  = 10 * time.Second
	confInterval = 10 * time.Second

	userAgent awsRequest.Option
	objects   map[string]*object
)

func InitFlags(fs *flag.FlagSet) {
	userAgentStr := "OIO-S3RT"

	fs.IntVar(&mpuSize, "s3rt.mpu_size", mpuSize, "[s3rt] MPU size")
	fs.IntVar(&size, "s3rt.size", size, "[s3rt] Put object size")
	fs.BoolVar(&ramBuffers, "s3rt.ram_buffers", ramBuffers, "[s3rt] Allocate data in RAM instead of using kernel devices")
	fs.BoolVar(&bucketRT, "s3rt.bucket_rt", bucketRT, "[s3rt] Perform a bucket roundtrip")
	fs.DurationVar(&confTimeout, "s3rt.timeout", confTimeout, "[s3rt] operation timeout")
	fs.DurationVar(&confInterval, "s3rt.interval", confInterval, "[s3rt] interval between polls")
	fs.StringVar(&userAgentStr, "s3rt.user_agent", userAgentStr, "[s3rt] userAgent identifier to add to requests")

	userAgent = awsRequest.WithAppendUserAgent(userAgentStr)
	objects = map[string]*object{
		"simple": &object{name: "s3roundtrip_object_simple", size: int64(size)},
		"ttfb":   &object{name: "s3roundtrip_ttfb_object", size: int64(1)},
		"mpu":    &object{name: "s3roundtrip_mpu_object", size: int64(mpuSize + 1)},
	}
}

type object struct {
	name string
	size int64
}

type FakeWriterAt struct {
	w io.Writer
}

func (fw FakeWriterAt) WriteAt(p []byte, offset int64) (n int, err error) {
	if buf, ok := fw.w.(*bytes.Buffer); ok {
		buf.Reset()
	}
	return fw.w.Write(p)
}

func max(a, b int) int {
	if a > b {
		return a
	}
	return b
}

func makeInputOutput(ramBuffers bool, size int) (io.ReadCloser, io.Writer, error) {
	if ramBuffers {
		buf := bytes.NewBuffer(make([]byte, size))
		return ioutil.NopCloser(buf), buf, nil
	}
	fd, err := os.Open("/dev/zero")
	if err != nil {
		return nil, nil, fmt.Errorf("while opening I/O device: %w", err)
	}
	return fd, ioutil.Discard, nil
}

var Factory = &s3rtDaemonFactory{}

type s3rtDaemonFactory struct{}

func (df *s3rtDaemonFactory) Daemons(inventory *common.Inventory) map[string]interface{} {
	params := make(map[string]interface{})
	namespace := inventory.Namespace

	services := inventory.Services["oioswift"]
	for _, service := range services {
		params[service.ID] = &s3rtDaemonParams{
			Namespace:   namespace,
			ServiceID:   service.ID,
			ServiceIP:   service.IP,
			ServicePort: service.Port,
		}
	}

	return params
}

func (df *s3rtDaemonFactory) New(paramsI interface{}) daemon.DaemonI {
	params := paramsI.(*s3rtDaemonParams)

	return &s3rtDaemon{
		namespace:   params.Namespace,
		serviceID:   params.ServiceID,
		serviceIP:   params.ServiceIP,
		servicePort: params.ServicePort,
	}
}

type s3rtDaemonParams struct {
	Namespace   string
	ServiceID   string
	ServiceIP   string
	ServicePort int
}

type s3rtDaemon struct {
	daemon.BaseDaemon

	namespace   string
	serviceID   string
	serviceIP   string
	servicePort int
	serviceAddr string
	bucket      string

	client        *s3c
	bucketCreated bool
}

type s3c struct {
	client s3iface.S3API
	src    io.ReadCloser
	dst    FakeWriterAt
	dl     s3manageriface.DownloaderAPI
	ul     s3manageriface.UploaderAPI
}

func (c *s3c) cleanup() error {
	return c.src.Close()
}

func (d *s3rtDaemon) Interval() time.Duration {
	return confInterval
}

func (d *s3rtDaemon) Init(ctx context.Context) error {
	serviceAddr := fmt.Sprintf("%s:%d", d.serviceIP, d.servicePort)
	endpoint := "http://" + serviceAddr
	config := aws.Config{
		Endpoint:         aws.String(endpoint),
		S3ForcePathStyle: aws.Bool(true),
		MaxRetries:       aws.Int(1),
		HTTPClient: &http.Client{
			Timeout: confTimeout,
			Transport: &http.Transport{
				// Prevent 100-Continue
				ExpectContinueTimeout: 0,
			},
		},
	}

	in, out, err := makeInputOutput(ramBuffers, max(mpuSize+1, size))
	if err != nil {
		return fmt.Errorf("while initializing the s3 client: %w", err)
	}

	// Ensure to load config and credentials from ~/.aws
	sess, err := session.NewSessionWithOptions(session.Options{
		Config:            config,
		SharedConfigState: session.SharedConfigEnable,
	})
	if err != nil {
		return fmt.Errorf("while initializing the s3 client: %w", err)
	}

	s3Client := s3.New(sess)
	d.client = &s3c{
		client: s3Client,
		src:    in,
		dst:    FakeWriterAt{w: out},
		dl: s3manager.NewDownloaderWithClient(s3Client, func(dl *s3manager.Downloader) {
			dl.RequestOptions = append(dl.RequestOptions, userAgent)
		}),
		ul: s3manager.NewUploaderWithClient(s3Client, func(ul *s3manager.Uploader) {
			ul.PartSize = int64(mpuSize)
			ul.RequestOptions = append(ul.RequestOptions, userAgent)
		}),
	}
	d.bucket = fmt.Sprintf("s3rt-%x", md5.Sum([]byte(endpoint)))
	d.serviceAddr = serviceAddr

	return nil
}

func (d *s3rtDaemon) Collect(ctx context.Context, m *common.MetricData) error {
	s3c := d.client

	var t time.Duration
	var err error

	if !d.bucketCreated || bucketRT {
		t, err = s3c.mb(ctx, d.bucket)
		if err == nil {
			d.bucketCreated = true
		} else if aerr, ok := err.(awserr.Error); ok && aerr.Code() == s3.ErrCodeBucketAlreadyExists {
			d.bucketCreated = true
		} else if !bucketRT {
			m.IncrError("s3rt", "s3rt")
			log.Warnln("error while precreating bucket", s3c, err)
		}
		if isDone(ctx) {
			return nil
		}
		if bucketRT {
			d.s3Publish(m, "mb", t, err, "")
		}
	}

	if isDone(ctx) {
		return nil
	}

	for type_, obj := range objects {
		t, err := s3c.put(ctx, d.bucket, obj)
		if err == nil {
			t, err = s3c.get(ctx, d.bucket, obj)
			d.s3Publish(m, "get", t, err, type_)
			t, err = s3c.del(ctx, d.bucket, obj)
			d.s3Publish(m, "del", t, err, type_)
		}
		d.s3Publish(m, "put", t, err, type_)

		if isDone(ctx) {
			return nil
		}
	}

	t, err = s3c.ls(ctx, d.bucket)
	d.s3Publish(m, "ls", t, err, "")
	if isDone(ctx) {
		return nil
	}

	if bucketRT {
		t, err = s3c.rb(ctx, d.bucket)
		d.s3Publish(m, "rb", t, err, "")
	}

	return nil
}

func (d *s3rtDaemon) Close() error {
	return d.client.cleanup()
}

func isDone(ctx context.Context) bool {
	select {
	case <-ctx.Done():
		return true
	default:
		return false
	}
}

func (d *s3rtDaemon) s3Publish(m *common.MetricData, request string, t time.Duration, err error, type_ string) {
	code := "2xx"
	if err != nil {
		if req, ok := err.(awserr.RequestFailure); ok {
			// Error is an AWS request failure
			status := req.StatusCode() / 100
			if status == 3 || status == 1 {
				// Treat redirects and continues as 2xx
				status = 2
			}
			code = strconv.Itoa(status) + "xx"
		} else {
			code = "other"
		}
	}
	labels := [][]string{
		[]string{"namespace", d.namespace},
		[]string{"service", d.serviceID},
		[]string{"service_type", "oioswift"},
		[]string{"service_addr", d.serviceAddr},
		[]string{"method", request},
		[]string{"code", code},
	}

	if type_ != "" {
		labels = append(labels, []string{"type", type_})
	}

	m.Set(&common.Metric{
		Name:   "s3rt_response_time_seconds",
		Labels: labels,
		Value:  float64(t.Seconds()),
	})
}

func (c *s3c) put(ctx context.Context, bucket string, obj *object) (time.Duration, error) {
	// We provision the writer with data at offset 0 so that the LimitedReader (which is the same as the writer)
	// always has enough data to upload
	if _, err := c.dst.WriteAt(make([]byte, obj.size), 0); err != nil {
		log.Warnln("Wailed to provision buffer with data")
	}
	input := &s3manager.UploadInput{
		Bucket: aws.String(bucket),
		Key:    aws.String(obj.name),
		Body:   io.LimitReader(c.src, obj.size),
	}
	requestCtx, cancel := context.WithTimeout(ctx, confTimeout)
	defer cancel()
	start := time.Now()
	_, err := c.ul.UploadWithContext(requestCtx, input)
	return time.Since(start), err
}

func (c *s3c) get(ctx context.Context, bucket string, obj *object) (time.Duration, error) {
	input := &s3.GetObjectInput{
		Bucket: aws.String(bucket),
		Key:    aws.String(obj.name),
	}
	requestCtx, cancel := context.WithTimeout(ctx, confTimeout)
	defer cancel()
	start := time.Now()
	_, err := c.dl.DownloadWithContext(requestCtx, c.dst, input)
	return time.Since(start), err
}

func (c *s3c) del(ctx context.Context, bucket string, obj *object) (time.Duration, error) {
	input := &s3.DeleteObjectInput{
		Bucket: aws.String(bucket),
		Key:    aws.String(obj.name),
	}
	requestCtx, cancel := context.WithTimeout(ctx, confTimeout)
	defer cancel()
	start := time.Now()
	_, err := c.client.DeleteObjectWithContext(requestCtx, input, userAgent)
	return time.Since(start), err
}

func (c *s3c) mb(ctx context.Context, bucket string) (time.Duration, error) {
	input := &s3.CreateBucketInput{
		Bucket: aws.String(bucket),
	}
	requestCtx, cancel := context.WithTimeout(ctx, confTimeout)
	defer cancel()
	start := time.Now()
	_, err := c.client.CreateBucketWithContext(requestCtx, input, userAgent)
	if aerr, ok := err.(awserr.Error); ok && aerr.Code() == s3.ErrCodeBucketAlreadyOwnedByYou {
		return time.Since(start), nil
	}
	return time.Since(start), err
}

func (c *s3c) rb(ctx context.Context, bucket string) (time.Duration, error) {
	input := &s3.DeleteBucketInput{
		Bucket: aws.String(bucket),
	}
	requestCtx, cancel := context.WithTimeout(ctx, confTimeout)
	defer cancel()
	start := time.Now()
	_, err := c.client.DeleteBucketWithContext(requestCtx, input, userAgent)
	return time.Since(start), err
}

func (c *s3c) ls(ctx context.Context, bucket string) (time.Duration, error) {
	input := &s3.ListObjectsInput{
		Bucket:  aws.String(bucket),
		MaxKeys: aws.Int64(1000),
	}
	requestCtx, cancel := context.WithTimeout(ctx, confTimeout)
	defer cancel()
	start := time.Now()
	_, err := c.client.ListObjectsWithContext(requestCtx, input, userAgent)
	return time.Since(start), err
}
